
import contextlib
import trio


@contextlib.asynccontextmanager
async def background_task(task, *args):
	async with trio.open_nursery() as nursery:
		nursery.start_soon(task, *args)
		yield
		nursery.cancel_scope.cancel()




import struct


class StreamOut:
	def __init__(self, endian):
		self.endian = endian
		self.data = bytearray()
		self.pos = 0
		self.stack = []
		
	def set_endian(self, endian): self.endian = endian
		
	def push(self): self.stack.append(self.pos)
	def pop(self): self.pos = self.stack.pop()
		
	def get(self): return bytes(self.data)
	def size(self): return len(self.data)
	def tell(self): return self.pos
	def seek(self, pos):
		if pos > len(self.data):
			self.data += bytes(pos - len(self.data))
		self.pos = pos
	def skip(self, num): self.seek(self.pos + num)
	def align(self, num): self.skip((num - self.pos % num) % num)
	def available(self): return len(self.data) - self.pos
	def eof(self): return self.pos >= len(self.data)
		
	def write(self, data):
		self.data[self.pos : self.pos + len(data)] = data
		self.pos += len(data)
		
	def pad(self, num, char=b"\0"):
		self.write(char * num)
		
	def ascii(self, data):
		self.write(data.encode("ascii"))
		
	def u8(self, value): self.write(bytes([value]))
	def u16(self, value): self.write(struct.pack(self.endian + "H", value))
	def u32(self, value): self.write(struct.pack(self.endian + "I", value))
	def u64(self, value): self.write(struct.pack(self.endian + "Q", value))
	
	def u16_be(self, value): self.write(struct.pack(">H", value))
	def u32_be(self, value): self.write(struct.pack(">I", value))
	
	def s8(self, value): self.write(struct.pack("b", value))
	def s16(self, value): self.write(struct.pack(self.endian + "h", value))
	def s32(self, value): self.write(struct.pack(self.endian + "i", value))
	def s64(self, value): self.write(struct.pack(self.endian + "q", value))
	
	def u24(self, value):
		if self.endian == ">": self.u24_be(value)
		else:
			self.u8(value & 0xFF)
			self.u16(value >> 8)
	
	def u24_be(self, value):
		self.u16(value >> 8)
		self.u8(value & 0xFF)
			
	def float(self, value): self.write(struct.pack(self.endian + "f", value))
	def double(self, value): self.write(struct.pack(self.endian + "d", value))
	
	def bool(self, value): self.u8(1 if value else 0)
	def char(self, value): self.u8(ord(value))
	def wchar(self, value): self.u16(ord(value))
	
	def chars(self, data): self.repeat(data, self.char)
	def wchars(self, data): self.repeat(data, self.wchar)
	
	def repeat(self, list, func):
		for value in list:
			func(value)
	

class StreamIn:
	def __init__(self, data, endian):
		self.endian = endian
		self.data = data
		self.pos = 0
		self.stack = []
	
	def set_endian(self, endian): self.endian = endian
	
	def push(self): self.stack.append(self.pos)
	def pop(self): self.pos = self.stack.pop()
		
	def get(self): return self.data
	def size(self): return len(self.data)
	
	def tell(self): return self.pos
	def seek(self, pos):
		if pos > self.size():
			raise OverflowError("Buffer overflow")
		self.pos = pos
	
	def skip(self, num): self.seek(self.pos + num)
	def align(self, num): self.skip((num - self.pos % num) % num)
	def eof(self): return self.pos == len(self.data)
	def available(self): return len(self.data) - self.pos
	
	def peek(self, num):
		if self.available() < num:
			raise OverflowError("Buffer overflow")
		return self.data[self.pos : self.pos + num]
		
	def read(self, num):
		data = self.peek(num)
		self.skip(num)
		return data
		
	def readall(self):
		return self.read(self.available())
		
	def pad(self, num, char=b"\0"):
		if self.read(num) != char * num:
			raise ValueError("Incorrect padding")
			
	def ascii(self, num):
		return self.read(num).decode("ascii")
		
	def u8(self): return self.read(1)[0]
	def u16(self): return struct.unpack(self.endian + "H", self.read(2))[0]
	def u32(self): return struct.unpack(self.endian + "I", self.read(4))[0]
	def u64(self): return struct.unpack(self.endian + "Q", self.read(8))[0]
	
	def u16_be(self): return struct.unpack(">H", self.read(2))[0]
	def u32_be(self): return struct.unpack(">I", self.read(4))[0]
	
	def s8(self): return struct.unpack("b", self.read(1))[0]
	def s16(self): return struct.unpack(self.endian + "h", self.read(2))[0]
	def s32(self): return struct.unpack(self.endian + "i", self.read(4))[0]
	def s64(self): return struct.unpack(self.endian + "q", self.read(8))[0]
	
	def u24(self):
		if self.endian == ">":
			return (self.u16() << 8) | self.u8()
		return self.u8() | (self.u16() << 8)
		
	def u24_be(self): return (self.u16() << 8) | self.u8()
	
	def float(self): return struct.unpack(self.endian + "f", self.read(4))[0]
	def double(self): return struct.unpack(self.endian + "d", self.read(8))[0]
	
	def bool(self): return bool(self.u8())
	def char(self): return chr(self.u8())
	def wchar(self): return chr(self.u16())
	
	def chars(self, num): return "".join(self.repeat(self.char, num))
	def wchars(self, num): return "".join(self.repeat(self.wchar, num))
	
	def repeat(self, func, count):
		return [func() for i in range(count)]




from Crypto.Cipher import AES
from ldn import streams, wlan, queue
from netlink import route
import contextlib
import secrets
import hashlib
import socket
import random
import struct
import math
import hmac
import copy
import trio


MACAddress = wlan.MACAddress


# Station accept policy
ACCEPT_ALL = 0
ACCEPT_NONE = 1
ACCEPT_BLACKLIST = 2
ACCEPT_WHITELIST = 3

# Authentication status code
AUTH_SUCCESS = 0
AUTH_DENIED_BY_POLICY = 1
AUTH_MALFORMED_REQUEST = 2
AUTH_INVALID_VERSION = 4
AUTH_UNEXPECTED = 5
AUTH_CHALLENGE_FAILURE = 6

# Disconnect reason
DISCONNECT_NETWORK_DESTROYED = 3
DISCONNECT_NETWORK_DESTROYED_FORCEFULLY = 4
DISCONNECT_STATION_REJECTED_BY_HOST = 5
DISCONNECT_CONNECTION_LOST = 6


AES_KEK_GENERATION_SOURCE = bytes.fromhex("4d870986c45d20722fba1053da92e8a9")
AES_KEY_GENERATION_SOURCE = bytes.fromhex("89615ee05c31b6805fe58f3da24f7aa8")
MASTER_KEY = bytes.fromhex("c2caaff089b9aed55694876055271c7d")

def decrypt_key(key, kek):
	aes = AES.new(kek, AES.MODE_ECB)
	return aes.decrypt(key)

def derive_key(inp, source):
	key = decrypt_key(AES_KEK_GENERATION_SOURCE, MASTER_KEY)
	key = decrypt_key(source, key)
	key = decrypt_key(AES_KEY_GENERATION_SOURCE, key)
	return decrypt_key(hashlib.sha256(inp).digest()[:16], key)

def generate_data_key(key, password):
	source = bytes.fromhex("f1e7018419a84f711da714c2cf919c9c")
	return derive_key(key + password.encode(), source)


class AuthenticationError(Exception):
	def __init__(self, status_code):
		self.status_code = status_code
	
	def __str__(self):
		return "Authentication failed with status %i" %self.status_code


class SessionInfo:
	def __init__(self):
		self.local_communication_id = None
		self.game_mode = None
		self.ssid = None
	
	def encode(self, endianness):
		stream = streams.StreamOut(endianness)
		stream.u64(self.local_communication_id)
		stream.pad(2)
		stream.u16(self.game_mode)
		stream.pad(4)
		stream.write(self.ssid)
		return stream.get()
	
	def decode(self, data, endianness):
		stream = streams.StreamIn(data, endianness)
		self.local_communication_id = stream.u64()
		stream.pad(2)
		self.game_mode = stream.u16()
		stream.pad(4)
		self.ssid = stream.read(16)


class ParticipantInfo:
	def __init__(self):
		self.reset()
	
	def reset(self):
		self.ip_address = "0.0.0.0"
		self.mac_address = MACAddress()
		self.connected = False
		self.name = ""
		self.app_version = 0
	
	def encode(self):
		name = self.name.encode()
		stream = streams.StreamOut(">")
		stream.write(socket.inet_aton(self.ip_address))
		stream.write(self.mac_address.encode())
		stream.bool(self.connected)
		stream.pad(1)
		stream.write(name + b"\0" * (32 - len(name)))
		stream.u16(self.app_version)
		stream.pad(10)
		return stream.get()
	
	def decode(self, data):
		stream = streams.StreamIn(data, ">")
		self.ip_address = socket.inet_ntoa(stream.read(4))
		self.mac_address = MACAddress(stream.read(6))
		self.connected = stream.bool()
		stream.pad(1)
		self.name = stream.read(32).rstrip(b"\0").decode()
		self.app_version = stream.u16()
		stream.pad(10)


class AdvertisementInfo:
	def __init__(self):
		self.key = None
		self.security_level = None
		self.station_accept_policy = None
		self.max_participants = None
		self.num_participants = None
		self.participants = None
		self.application_data = None
		self.challenge = None
	
	def encode(self):
		stream = streams.StreamOut(">")
		stream.write(self.key)
		stream.u16(self.security_level)
		stream.u8(self.station_accept_policy)
		stream.pad(3)
		stream.u8(self.max_participants)
		stream.u8(self.num_participants)
		for participant in self.participants:
			stream.write(participant.encode())
		stream.pad(2)
		stream.u16(len(self.application_data))
		stream.write(self.application_data + b"\0" * (384 - len(self.application_data)))
		stream.pad(412)
		stream.u64(self.challenge)
		return stream.get()
	
	def decode(self, data):
		stream = streams.StreamIn(data, ">")
		
		self.key = stream.read(16)
		self.security_level = stream.u16()
		self.station_accept_policy = stream.u8()
		stream.pad(3)
		self.max_participants = stream.u8()
		self.num_participants = stream.u8()
		
		self.participants = []
		for i in range(8):
			participant = ParticipantInfo()
			participant.decode(stream.read(56))
			self.participants.append(participant)
		
		stream.pad(2)
		
		beacon_size = stream.u16()
		beacon_data = stream.read(384)
		self.application_data = beacon_data[:beacon_size]
		
		stream.pad(412)
		
		self.challenge = stream.u64()


class AdvertisementFrame:
	def __init__(self):
		self.header = None
		self.version = None
		self.encryption = None
		self.nonce = None
		self.info = None
	
	def encrypt(self, data):
		if self.encryption == 1:
			return data
		
		source = bytes.fromhex("191884743e24c77d87c69e4207d0c438")
		key = derive_key(self.header.encode(">"), source)
		aes = AES.new(key, AES.MODE_CTR, nonce=self.nonce)
		return aes.encrypt(data)
	
	def decrypt(self, data):
		if self.encryption == 1:
			return data
		
		source = bytes.fromhex("191884743e24c77d87c69e4207d0c438")
		key = derive_key(self.header.encode(">"), source)
		aes = AES.new(key, AES.MODE_CTR, nonce=self.nonce)
		return aes.decrypt(data)
	
	def encode(self):
		stream = streams.StreamOut(">")
		stream.u8(0x7F) # Vendor-specific
		stream.u24(0x0022AA) # Nintendo
		stream.u8(4) # LDN
		stream.pad(1)
		stream.u16(0x101) # Advertisement frame
		stream.pad(4)
		
		substream = streams.StreamOut(">")
		substream.write(self.header.encode(">"))
		substream.u8(self.version)
		substream.u8(self.encryption)
		substream.u16(0x500)
		substream.write(self.nonce)
		header = substream.get()
		
		info = self.info.encode()
		message = header + bytes(32) + info
		sha = hashlib.sha256(message).digest()
		
		stream.write(header)
		stream.write(self.encrypt(sha + info))
		return stream.get()
	
	def decode(self, data):
		stream = streams.StreamIn(data, ">")
		if stream.u8() != 0x7F:
			raise ValueError("Action frame is not vendor-specific")
		if stream.u24() != 0x0022AA:
			raise ValueError("Action frame has wrong OUI")
		
		if stream.u8() != 4:
			raise ValueError("Action frame is not for LDN")
		
		stream.pad(1)
		if stream.u16() != 0x101:
			raise ValueError("Action frame is not an advertisement frame")
		stream.pad(4)
		
		header = stream.peek(0x28)
		
		self.header = SessionInfo()
		self.header.decode(stream.read(32), ">")
		
		self.version = stream.u8()
		if self.version not in [2, 3]:
			raise ValueError("Advertisement frame has unsupported version number")
		
		self.encryption = stream.u8()
		if self.encryption not in [1, 2]:
			raise ValueError("Advertisement frame has invalid encryption algorithm")
		
		size = stream.u16()
		if size != 0x500:
			raise ValueError("Advertisement frame has unexpected size field")
		
		self.nonce = stream.read(4)
		
		body = self.decrypt(stream.read(32 + size))
		sha = body[:32]
		info = body[32:]
		
		message = header + bytes(32) + info
		if hashlib.sha256(message).digest() != sha:
			raise ValueError("Advertisement frame has wrong SHA-256 hash")
		
		self.info = AdvertisementInfo()
		self.info.decode(info)
		
		
class ChallengeRequest:
	def __init__(self):
		self.token = None
		self.nonce = None
		self.device_id = None
		self.params1 = []
		self.params2 = []
	
	def encode(self):
		stream = streams.StreamOut("<")
		stream.u8(0) # Always 0
		stream.u8(0) # Always 0
		stream.u8(len(self.params1))
		stream.u8(len(self.params2))
		stream.u8(0) # Debug check
		stream.pad(3)
		
		stream.u64(self.token)
		stream.u64(self.nonce)
		stream.u64(self.device_id)
		
		stream.pad(0x70)
		
		stream.repeat(self.params1, stream.u64)
		stream.pad(8 * (8 - len(self.params1)))
		
		stream.repeat(self.params2, stream.u64)
		stream.pad(8 * (64 - len(self.params2)))
		
		body = stream.get()
		
		key = bytes.fromhex("f84b487fb37251c263bf11609036589266af70ca79b44c93c7370c5769c0f602")
		mac = hmac.digest(key, body, "sha256")
		
		stream = streams.StreamOut("<")
		stream.u32(0)
		stream.write(mac)
		stream.pad(12)
		stream.write(body)
		
		return stream.get()
	
	def decode(self, data):
		if len(data) != 0x300:
			raise ValueError("Challenge request has wrong size")
		
		stream = streams.StreamIn(data, "<")
		stream.pad(4)
		mac = stream.read(32)
		stream.pad(12)
		body = stream.read(0x2D0)
		
		key = bytes.fromhex("f84b487fb37251c263bf11609036589266af70ca79b44c93c7370c5769c0f602")
		if mac != hmac.digest(key, body, "sha256"):
			raise ValueError("Challenge request has wrong HMAC")
		
		stream = streams.StreamIn(body, "<")
		stream.pad(2)
		n1 = stream.u8()
		n2 = stream.u8()
		stream.pad(4)
		
		self.token = stream.u64()
		self.nonce = stream.u64()
		self.device_id = stream.u64()
		
		stream.pad(0x70)
		
		self.params1 = stream.repeat(stream.u64, 8)[:n1]
		self.params2 = stream.repeat(stream.u64, 8)[:n2]


class ChallengeResponse:
	def __init__(self):
		self.nonce = None
		self.device_id = None
		self.device_id_host = None
	
	def encode(self):
		stream = streams.StreamOut("<")
		stream.u8(0) # Always 0
		stream.u8(0) # Always 0
		stream.pad(6)
		stream.u64(self.nonce)
		stream.u64(self.device_id)
		stream.u64(self.device_id_host)
		stream.pad(0xB0)
		
		body = stream.get()
	
		key = bytes.fromhex("f84b487fb37251c263bf11609036589266af70ca79b44c93c7370c5769c0f602")
		mac = hmac.digest(key, body, "sha256")
		
		stream = streams.StreamOut("<")
		stream.u32(0)
		stream.write(mac)
		stream.pad(12)
		stream.write(body)
		return stream.get()
	
	def decode(self, data):
		if len(data) != 0x100:
			raise ValueError("Challenge response has wrong size")
		
		stream = streams.StreamIn(data, "<")
		stream.pad(4)
		mac = stream.read(32)
		stream.pad(12)
		body = stream.read(0xD0)
		
		key = bytes.fromhex("f84b487fb37251c263bf11609036589266af70ca79b44c93c7370c5769c0f602")
		if mac != hmac.digest(key, body, "sha256"):
			raise ValueError("Challenge response has wrong HMAC")
		
		stream = streams.StreamIn(body, "<")
		stream.pad(8)
		self.nonce = stream.u64()
		self.device_id = stream.u64()
		self.device_id_host = stream.u64()
		stream.pad(0xB0)


class AuthenticationRequest:
	def __init__(self):
		self.username = None
		self.app_version = None
		self.challenge = None
	
	def encode(self, version):
		stream = streams.StreamOut(">")
		
		name = self.username.encode()
		stream.write(name + b"\0" * (32 - len(name)))
		stream.u16(self.app_version)
		stream.pad(30)
		
		if version >= 3:
			stream.pad(0x24)
			if self.challenge is not None:
				stream.write(self.challenge)
		return stream.get()
	
	def decode(self, data, version):
		stream = streams.StreamIn(data, ">")
		
		self.username = stream.read(32).rstrip(b"\0").decode()
		self.app_version = stream.u16()
		stream.pad(30)
		
		if version >= 3:
			stream.pad(0x24)
			if not stream.eof():
				self.challenge = stream.read(0x300)


class AuthenticationResponse:
	def __init__(self):
		self.challenge = None
	
	def encode(self, version):
		stream = streams.StreamOut(">")
		if version >= 3:
			stream.pad(0x84)
			if self.challenge is not None:
				stream.write(self.challenge)
		return stream.get()
	
	def decode(self, data, version):
		stream = streams.StreamIn(data, ">")
		if version >= 3:
			stream.pad(0x84)
			if not stream.eof():
				self.challenge = stream.read(0x100)


class AuthenticationFrame:
	def __init__(self):
		self.version = None
		self.status_code = None
		self.header = None
		self.network_key = None
		self.authentication_key = None
		self.payload = None
	
	def encode(self):
		payload = self.payload.encode(self.version)
		
		stream = streams.StreamOut(">")
		stream.u24(0x0022AA) # Nintendo
		stream.u16(0x102) # Authentication frame
		stream.pad(1)
		
		stream.u8(self.version)
		stream.u8(len(payload) & 0xFF)
		stream.u8(self.status_code)
		stream.u8(isinstance(self.payload, AuthenticationResponse))
		stream.u8(len(payload) >> 8)
		stream.pad(3)
		
		stream.write(self.header.encode("<"))
		stream.write(self.network_key)
		stream.write(self.authentication_key)
		stream.write(payload)
		return stream.get()
	
	def decode(self, data):
		stream = streams.StreamIn(data, ">")
		if stream.u24() != 0x0022AA:
			raise ValueError("Data frame has wrong OUI")
		if stream.u16() != 0x102:
			raise ValueError("Data frame is not an authentication frame")
		stream.pad(1)
		
		self.version = stream.u8()
		size_lo = stream.u8()
		self.status_code = stream.u8()
		is_response = stream.u8()
		size_hi = stream.u8()
		stream.pad(3)
		
		self.header = SessionInfo()
		self.header.decode(stream.read(32), "<")
		self.network_key = stream.read(16)
		self.authentication_key = stream.read(16)
		
		size = (size_hi << 8) | size_lo
		if stream.available() != size:
			raise ValueError("Authentication frame has wrong size")
		
		if is_response:
			self.payload = AuthenticationResponse()
		else:
			self.payload = AuthenticationRequest()
		self.payload.decode(stream.read(size), self.version)


class DisconnectFrame:
	def __init__(self):
		self.reason = None
	
	def encode(self):
		stream = streams.StreamOut(">")
		stream.u24(0x0022AA) # Nintendo
		stream.u16(0x103) # Disconnect frame
		stream.pad(1)
		
		stream.u8(self.reason)
		stream.pad(31)
		return stream.get()
	
	def decode(self, data):
		stream = streams.StreamIn(data, ">")
		if stream.u24() != 0x0022AA:
			raise ValueError("Data frame has wrong OUI")
		if stream.u16() != 0x103:
			raise ValueError("Data frame is not a disconnect frame")
		stream.pad(1)
		
		self.reason = stream.u8()
		stream.pad(31)

		
class NetworkInfo:
	def __init__(self):
		self.address = None
		self.channel = None
		
		self.local_communication_id = None
		self.game_mode = None
		self.ssid = None
		self.version = None
		self.key = None
		self.security_level = None
		self.accept_policy = None
		self.max_participants = None
		self.num_participants = None
		self.participants = None
		self.application_data = None
		self.challenge = None
		self.nonce = None
	
	def check(self, info):
		if self.address != info.address: return False
		if self.channel != info.channel: return False
		
		if info.local_communication_id != self.local_communication_id: return False
		if info.game_mode != self.game_mode: return False
		if info.ssid != self.ssid: return False
		if info.version != self.version: return False
		if info.security_level != self.security_level: return False
		return True
	
	def parse(self, frame):
		self.local_communication_id = frame.header.local_communication_id
		self.game_mode = frame.header.game_mode
		self.ssid = frame.header.ssid
		self.version = frame.version
		self.key = frame.info.key
		self.security_level = frame.info.security_level
		self.accept_policy = frame.info.station_accept_policy
		self.max_participants = frame.info.max_participants
		self.num_participants = frame.info.num_participants
		self.participants = frame.info.participants
		self.application_data = frame.info.application_data
		self.challenge = frame.info.challenge
		self.nonce = frame.nonce
	
	def build(self):
		header = SessionInfo()
		header.local_communication_id = self.local_communication_id
		header.game_mode = self.game_mode
		header.ssid = self.ssid
		
		info = AdvertisementInfo()
		info.key = self.key
		info.security_level = self.security_level
		info.station_accept_policy = self.accept_policy
		info.max_participants = self.max_participants
		info.num_participants = self.num_participants
		info.participants = self.participants
		info.application_data = self.application_data
		info.challenge = self.challenge
		
		frame = AdvertisementFrame()
		frame.header = header
		frame.version = self.version
		frame.encryption = 1 if self.security_level == 3 else 2
		frame.nonce = self.nonce
		frame.info = info
		return frame


class ConnectNetworkParam:
	def __init__(self):
		self.ifname = "ldn"
		self.phyname = "phy0"
		
		self.network = None
		self.password = ""
		
		self.name = ""
		self.app_version = 0
		
		self.enable_challenge = True
		self.device_id = random.randint(0, 0xFFFFFFFFFFFFFFFF)

	def check(self):
		if self.network is None: raise ValueError("network is required")
		if self.network.version not in [2, 3]:
			raise ValueError("Network version not supported")


class CreateNetworkParam:
	def __init__(self):
		self.ifname = "ldn"
		self.ifname_monitor = "ldn-mon"
		self.phyname = "phy0"
		self.phyname_monitor = "phy0"
		
		self.local_communication_id = None
		self.game_mode = None
		
		self.max_participants = 8
		self.application_data = b""
		self.accept_policy = ACCEPT_ALL
		self.accept_filter = []
		self.security_level = 1
		self.ssid = None
		
		self.name = ""
		self.app_version = 0
		
		self.channel = None
		self.key = None
		self.password = ""
		
		self.version = 3
		self.enable_challenge = True
		self.device_id = random.randint(0, 0xFFFFFFFFFFFFFFFF)

	def check(self):
		if self.local_communication_id is None: raise ValueError("local_communication_id is required")
		if self.game_mode is None: raise ValueError("game_mode is required")
		if self.max_participants > 8: raise ValueError("max_participants is too high")
		if len(self.application_data) > 0x180: raise ValueError("application_data is too large")
		if self.ssid is not None and len(self.ssid) != 16:
			raise ValueError("ssid has wrong size")
		if self.channel is not None and not wlan.is_valid_channel(self.channel):
			raise ValueError("channel is invalid")
		if self.key is not None and len(self.key) != 16:
			raise ValueError("key has wrong size")
		if self.version not in [2, 3]:
			raise ValueError("version is invalid")


class DisconnectEvent:
	def __init__(self, reason):
		self.reason = reason

class JoinEvent:
	def __init__(self, index, participant):
		self.index = index
		self.participant = participant

class LeaveEvent:
	def __init__(self, index, participant):
		self.index = index
		self.participant = participant

class AcceptPolicyChanged:
	def __init__(self, old, new):
		self.old = old
		self.new = new

class ApplicationDataChanged:
	def __init__(self, old, new):
		self.old = old
		self.new = new


class AdvertisementMonitor:
	def __init__(self, monitor):
		self.monitor = monitor
	
	async def receive(self):
		# Vendor-specific, Nintendo OUI, LDN, Advertisement
		header = bytes([0x7F, 0x00, 0x22, 0xAA, 0x04, 0x00, 0x01, 0x01])
		while True:
			# Receive a single frame
			radiotap = await self.monitor.recv()
			
			# Check if we received an action frame
			if len(radiotap.data) < 2 or struct.unpack_from(">H", radiotap.data)[0] != 0xD000:
				continue
			
			action = wlan.ActionFrame()
			try: action.decode(radiotap.data)
			except Exception:
				continue # Skip invalid frames
			
			# Check if we received an advertisement frame from LDN
			if not action.action.startswith(header):
				continue
			
			# Decode the frame itself
			frame = AdvertisementFrame()
			try: frame.decode(action.action)
			except Exception:
				continue # Skip invalid frames
			
			info = NetworkInfo()
			info.address = action.source
			info.channel = wlan.map_frequency(radiotap.frequency)
			info.parse(frame)
			return info

	async def scan(self, channels, dwell_time):
		networks = []
		async def scan_frames():
			addresses = []
			while True:
				network = await self.receive()
				if network.address not in addresses:
					addresses.append(network.address)
					networks.append(network)
		
		async with util.background_task(scan_frames):
			for channel in channels:
				await self.monitor.set_channel(channel)
				await trio.sleep(dwell_time)
		return networks

		
class STANetwork:
	def __init__(self, interface, router, param):
		self.interface = interface
		self.router = router
		
		self.network = param.network
		self.param = param
		
		self.authentication_key = secrets.token_bytes(16)
		
		self.network_id = None
		self.participant_id = None
		
		self.events = queue.create()
		self.advertisements = queue.create()
	
	def check_authentication_response(self, address, data):
		if address != self.network.address: return False
		
		frame = AuthenticationFrame()
		try:
			frame.decode(data)
		except Exception:
			return False # Ignore invalid frames
		
		if not isinstance(frame.payload, AuthenticationResponse): return False
		
		if frame.header.local_communication_id != self.network.local_communication_id: return False
		if frame.header.game_mode != self.network.game_mode: return False
		if frame.header.ssid != self.network.ssid: return False
		
		if frame.network_key != self.network.key: return False
		if frame.authentication_key != self.authentication_key: return False
		
		if frame.status_code != 0:
			raise AuthenticationError(frame.status_code)
		return True
	
	def info(self):
		return self.network
	
	def participant(self):
		return self.network.participants[self.participant_id]
	
	async def next_event(self):
		return await self.events.get()
	
	@contextlib.asynccontextmanager
	async def start(self):
		await self.authenticate()
		async with util.background_task(self.process_events):
			await self.initialize_network()
			async with util.background_task(self.monitor_network):
				yield
	
	async def process_events(self):
		while True:
			event = await self.interface.next_event()
			if isinstance(event, wlan.FrameEvent):
				if event.frame.source != self.network.address:
					continue # Only process frames from the host
				
				frame = AdvertisementFrame()
				try: frame.decode(event.frame.action)
				except Exception:
					continue # Skip invalid frames
				
				info = NetworkInfo()
				info.address = event.frame.source
				info.channel = wlan.map_frequency(event.frequency)
				info.parse(frame)
				if not self.network.check(info):
					raise ConnectionError("Received incompatible advertisement frame from host")
				await self.advertisements.put(info)
			elif isinstance(event, wlan.DataFrameEvent):
				frame = DisconnectFrame()
				frame.decode(event.data)
				await self.events.put(DisconnectEvent(frame.reason))
			elif isinstance(event, wlan.DisassociationEvent):
				await self.events.put(DisconnectEvent(DISCONNECT_CONNECTION_LOST))
	
	async def authenticate(self):
		request = AuthenticationRequest()
		request.username = self.param.name
		request.app_version = self.param.app_version
		
		if self.param.enable_challenge:
			challenge = ChallengeRequest()
			challenge.token = self.network.challenge
			challenge.nonce = random.randint(0, 0xFFFFFFFFFFFFFFFF)
			challenge.device_id = self.param.device_id
			request.challenge = challenge.encode()
		
		header = SessionInfo()
		header.local_communication_id = self.network.local_communication_id
		header.game_mode = self.network.game_mode
		header.ssid = self.network.ssid
		
		frame = AuthenticationFrame()
		frame.version = self.network.version
		frame.status_code = 0
		frame.header = header
		frame.network_key = self.network.key
		frame.authentication_key = self.authentication_key
		frame.payload = request
		
		# Attempt authentication up to three times
		for i in range(3):
			await self.interface.send_data_frame(self.network.address, frame.encode())
			
			# Resend the authentication request if we do not
			# receive a response after 700 milliseconds
			with trio.move_on_after(.7):
				while True:
					event = await self.interface.next_event()
					if isinstance(event, wlan.DataFrameEvent):
						if self.check_authentication_response(event.address, event.data):
							return
					elif isinstance(event, wlan.DisassociationEvent):
						raise ConnectionError("Station was disassociated")
		raise ConnectionError("Authentication timeout (password may be wrong)")
	
	async def wait_for_network(self):
		while True:
			network = await self.advertisements.get()
			for index, participant in enumerate(network.participants):
				if participant.mac_address == self.interface.address:
					return network, index
	
	async def initialize_network(self):
		await self.interface.set_authorized()
		
		# Wait until the host has updated the advertisement frame
		network = None
		with trio.move_on_after(1):
			network, index = await self.wait_for_network()
		
		if network is None:
			raise ConnectionError("Failed to obtain IP address after joining network (timeout)")
		
		# Initialize local state
		self.network = network
		self.network_id = int(network.participants[0].ip_address.split(".")[2])
		self.participant_id = index
		
		# Initialize interface address
		attrs = {
			route.IFA_LOCAL: socket.inet_aton(network.participants[index].ip_address),
			route.IFA_BROADCAST: socket.inet_aton("169.254.%i.255" %self.network_id)
		}
		await self.router.add_address(
			socket.AF_INET, 24, route.IFA_F_PERMANENT, route.RT_SCOPE_UNIVERSE,
			self.interface.index, attrs
		)
		
		# Create a static neighbor entry for each participant
		for participant in network.participants:
			if participant.connected:
				attrs = {
					route.NDA_DST: socket.inet_aton(participant.ip_address),
					route.NDA_LLADDR: participant.mac_address.encode()
				}
				await self.router.add_neighbor(socket.AF_INET, self.interface.index, route.NUD_PERMANENT, 0, 0, attrs)
	
	async def monitor_network(self):
		# Monitors advertisement frames to get
		# notified when the network changes
		while True:
			network = await self.advertisements.get()
			
			# Check if the accept policy has changed
			if network.accept_policy != self.network.accept_policy:
				await self.events.put(AcceptPolicyChanged(self.network.accept_policy, network.accept_policy))
			
			# Check if the application data has changed
			if network.application_data != self.network.application_data:
				await self.events.put(ApplicationDataChanged(self.network.application_data, network.application_data))
			
			# Remove participants that are gone
			for i in range(8):
				old = self.network.participants[i]
				new = network.participants[i]
				if old.connected and old.mac_address != new.mac_address:
					attrs = {
						route.NDA_DST: socket.inet_aton(old.ip_address),
						route.NDA_LLADDR: old.mac_address.encode()
					}
					await self.router.remove_neighbor(socket.AF_INET, self.interface.index, route.NUD_PERMANENT, 0, 0, attrs)
					await self.events.put(LeaveEvent(i, old))
			
			# Register new participants
			for i in range(8):
				old = self.network.participants[i]
				new = network.participants[i]
				if new.connected and old.mac_address != new.mac_address:
					attrs = {
						route.NDA_DST: socket.inet_aton(new.ip_address),
						route.NDA_LLADDR: new.mac_address.encode()
					}
					await self.router.add_neighbor(socket.AF_INET, self.interface.index, route.NUD_PERMANENT, 0, 0, attrs)
					await self.events.put(JoinEvent(i, new))
			
			# Update local state
			self.network = network


class APNetwork:
	def __init__(self, interface, monitor, router, param):
		self.interface = interface
		self.monitor = monitor
		self.router = router
		
		self.accept_filter = param.accept_filter
		self.enable_challenge = param.enable_challenge
		self.device_id = param.device_id
		
		self.nonce = random.randint(0, 0xFFFFFFFF)
		self.network_id = random.randint(1, 127)
		
		participant = ParticipantInfo()
		participant.ip_address = "169.254.%i.1" %self.network_id
		participant.mac_address = interface.address
		participant.connected = True
		participant.name = param.name
		participant.app_version = param.app_version
		
		participants = [participant]
		for i in range(7):
			participants.append(ParticipantInfo())
		
		self.network = NetworkInfo()
		self.network.address = interface.address
		self.network.channel = param.channel
		self.network.local_communication_id = param.local_communication_id
		self.network.game_mode = param.game_mode
		self.network.ssid = param.ssid
		self.network.version = param.version
		self.network.key = param.key
		self.network.security_level = param.security_level
		self.network.accept_policy = param.accept_policy
		self.network.max_participants = param.max_participants
		self.network.num_participants = 1
		self.network.participants = participants
		self.network.application_data = param.application_data
		self.network.challenge = random.randint(0, 0xFFFFFFFFFFFFFFFF)
		self.network.nonce = struct.pack(">I", self.nonce)
		
		self.events = queue.create()
	
	def make_authentication_response(self, status, version, key, challenge=b""):
		header = SessionInfo()
		header.local_communication_id = self.network.local_communication_id
		header.game_mode = self.network.game_mode
		header.ssid = self.network.ssid
		
		response = AuthenticationResponse()
		response.challenge = challenge
		
		frame = AuthenticationFrame()
		frame.version = version
		frame.status_code = status
		frame.header = header
		frame.network_key = self.network.key
		frame.authentication_key = key
		frame.payload = response
		return frame
	
	def check_accept_policy(self, address):
		if self.network.accept_policy == ACCEPT_ALL: return True
		if self.network.accept_policy == ACCEPT_NONE: return False
		if self.network.accept_policy == ACCEPT_BLACKLIST:
			return address not in self.accept_filter
		if self.network.accept_policy == ACCEPT_WHITELIST:
			return address in self.accept_filter
		return False
	
	def check_authentication_request(self, address, frame):
		if frame.version not in [2, 3]: return AUTH_INVALID_VERSION
		
		if frame.status_code != 0: return AUTH_MALFORMED_REQUEST
		if frame.header.local_communication_id != self.network.local_communication_id: return AUTH_MALFORMED_REQUEST
		if frame.header.game_mode != self.network.game_mode: return AUTH_MALFORMED_REQUEST
		if frame.header.ssid != self.network.ssid: return AUTH_MALFORMED_REQUEST
		if frame.network_key != self.network.key: return AUTH_MALFORMED_REQUEST
		if not isinstance(frame.payload, AuthenticationRequest): return AUTH_MALFORMED_REQUEST
		
		if not self.check_accept_policy(address):
			return AUTH_DENIED_BY_POLICY
	
	def process_authentication_challenge(self, challenge):
		if not self.enable_challenge: return b""
		
		request = ChallengeRequest()
		try:
			request.decode(challenge)
		except Exception:
			return None
		
		if request.token != self.network.challenge:
			return None
		
		response = ChallengeResponse()
		response.nonce = request.nonce
		response.device_id = request.device_id
		response.device_id_host = self.device_id
		return response.encode()
	
	def update_nonce(self):
		self.nonce = (self.nonce + 1) & 0xFFFFFFFF
		self.network.nonce = struct.pack(">I", self.nonce)
	
	def info(self):
		return self.network
	
	def participant(self):
		return self.network.participants[0]
	
	def set_application_data(self, data):
		self.network.application_data = data
		self.update_nonce()
	
	def set_accept_policy(self, policy):
		self.network.accept_policy = policy
		self.update_nonce()
	
	def set_accept_filter(self, filter):
		self.accept_filter = filter
		
	async def kick(self, index):
		participant = self.network.participants[index]
		if participant.connected:
			frame = DisconnectFrame()
			frame.reason = DISCONNECT_STATION_REJECTED_BY_HOST
			await self.interface.send_data_frame(participant.mac_address, frame.encode())
			await self.interface.remove_station(participant.mac_address)
			await self.process_disassociation(participant.mac_address)
	
	async def next_event(self):
		return await self.events.get()
	
	@contextlib.asynccontextmanager
	async def start(self):
		await self.initialize_network()
		async with util.background_task(self.process_events):
			async with util.background_task(self.send_advertisements):
				yield
				await self.destroy_network()
	
	async def process_events(self):
		while True:
			event = await self.interface.next_event()
			if isinstance(event, wlan.DataFrameEvent):
				response = await self.process_authentication_event(event)
				await self.interface.send_data_frame(event.address, response.encode())
			elif isinstance(event, wlan.DisassociationEvent):
				await self.process_disassociation(event.address)
	
	async def process_authentication_event(self, event):
		frame = AuthenticationFrame()
		
		try:
			frame.decode(event.data)
		except Exception:
			return self.make_authentication_response(AUTH_MALFORMED_REQUEST, self.network.version, bytes(16))
		
		error = self.check_authentication_request(event.address, frame)
		if error is not None:
			return self.make_authentication_response(error, self.network.version, frame.authentication_key)
		
		challenge = self.process_authentication_challenge(frame.payload.challenge)
		if challenge is None:
			return self.make_authentication_response(AUTH_CHALLENGE_FAILURE, self.network.version, frame.authentication_key)
		
		await self.register_participant(event.address, frame.payload.username, frame.payload.app_version)
		
		return self.make_authentication_response(AUTH_SUCCESS, self.network.version, frame.authentication_key, challenge)
	
	async def register_participant(self, address, name, app_version):
		# Allocate an ip address
		for index in range(8):
			if not self.network.participants[index].connected:
				break
		
		participant = ParticipantInfo()
		participant.ip_address = "169.254.%i.%i" %(self.network_id, (index + 1))
		participant.mac_address = address
		participant.connected = True
		participant.name = name
		participant.app_version = app_version
		
		self.network.participants[index] = participant
		self.update_nonce()

		await self.interface.set_authorized(address)
		
		# Add neighbor entry
		attrs = {
			route.NDA_DST: socket.inet_aton(participant.ip_address),
			route.NDA_LLADDR: participant.mac_address.encode()
		}
		await self.router.add_neighbor(socket.AF_INET, self.interface.index, route.NUD_PERMANENT, 0, 0, attrs)
		
		await self.events.put(JoinEvent(index, participant))
	
	async def process_disassociation(self, address):
		for index, participant in enumerate(self.network.participants):
			if participant.connected and participant.mac_address == address:
				break
		else: return
		
		participant.connected = False
		self.update_nonce()
		
		# Remove neighbor entry
		attrs = {
			route.NDA_DST: socket.inet_aton(participant.ip_address),
			route.NDA_LLADDR: participant.mac_address.encode()
		}
		await self.router.remove_neighbor(socket.AF_INET, self.interface.index, route.NUD_PERMANENT, 0, 0, attrs)
		
		await self.events.put(LeaveEvent(index, participant))
	
	async def send_advertisements(self):
		while True:
			await self.send_advertisement()
			await trio.sleep(.1)
	
	async def send_advertisement(self):
		frame = self.network.build()
		
		action = wlan.ActionFrame()
		action.source = self.interface.address
		action.action = frame.encode()
		
		radiotap = wlan.RadiotapFrame()
		radiotap.data = action.encode()
		await self.monitor.send(radiotap)
	
	async def initialize_network(self):
		host = self.network.participants[0]

		attrs = {
			route.IFA_LOCAL: socket.inet_aton(host.ip_address),
			route.IFA_BROADCAST: socket.inet_aton("169.254.%i.255" %self.network_id)
		}
		await self.router.add_address(
			socket.AF_INET, 24, route.IFA_F_PERMANENT, route.RT_SCOPE_UNIVERSE,
			self.interface.index, attrs
		)

		attrs = {
			route.NDA_DST: socket.inet_aton(host.ip_address),
			route.NDA_LLADDR: host.mac_address.encode()
		}
		await self.router.add_neighbor(socket.AF_INET, self.interface.index, route.NUD_PERMANENT, 0, 0, attrs)
	
	async def destroy_network(self):
		for participant in self.network.participants:
			if participant.connected:
				frame = DisconnectFrame()
				frame.reason = DISCONNECT_NETWORK_DESTROYED
				await self.interface.send_data_frame(participant.mac_address, frame.encode())


async def scan(ifname="ldn", phyname="phy0", channels=[1, 6, 11], dwell_time=.110):
	if not channels: return []

	# Check if all channels are valid
	for channel in channels:
		if not wlan.is_valid_channel(channel):
			raise ValueError("Invalid channel: %i" %channel)
	
	async with wlan.create() as factory:
		async with factory.create_monitor(phyname, ifname) as monitor:
			scanner = AdvertisementMonitor(monitor)
			return await scanner.scan(channels, dwell_time)

@contextlib.asynccontextmanager
async def connect(param):
	param = copy.copy(param)
	param.check()
	
	network = param.network
	
	key = None
	if network.security_level == 1:
		key = generate_data_key(network.key, param.password)
	
	async with wlan.create() as factory:
		async with factory.connect_network(param.phyname, param.ifname, network.ssid.hex(), network.channel, key) as interface:
			async with route.connect() as router:
				network = STANetwork(interface, router, param)
				async with network.start():
					yield network

@contextlib.asynccontextmanager
async def create_network(param):
	param = copy.copy(param)
	if param.ssid is None: param.ssid = secrets.token_bytes(16)
	if param.channel is None: param.channel = random.choice([1, 6, 11])
	if param.key is None: param.key = secrets.token_bytes(16)
	param.check()
	
	key = None
	if param.security_level == 1:
		key = generate_data_key(param.key, param.password)
	
	async with wlan.create() as factory:
		async with factory.create_monitor(param.phyname_monitor, param.ifname_monitor) as monitor:
			await monitor.set_channel(param.channel)
			async with factory.create_network(param.phyname, param.ifname, param.ssid.hex(), param.channel, key, param.max_participants) as interface:
				async with route.connect() as router:
					network = APNetwork(interface, monitor, router, param)
					async with network.start():
						yield network




# This file implements WLAN functions using NL80211

from netlink import generic, nl80211
from ldn import streams, util, queue
import contextlib
import netlink
import string
import struct
import socket
import fcntl
import trio


SIOCGIFFLAGS = 0x8913
SIOCSIFFLAGS = 0x8914

IFF_UP = 1

ETH_P_ALL = 3
ETH_P_OUI = 0x88B7


IEEE80211_FTYPE_MGMT = 0
IEEE80211_FTYPE_CTL = 4
IEEE80211_FTYPE_DATA = 8
IEEE80211_FTYPE_EXT = 0xC

IEEE80211_STYPE_ASSOC_REQ = 0
IEEE80211_STYPE_ASSOC_RESP = 0x10
IEEE80211_STYPE_REASSOC_REQ = 0x20
IEEE80211_STYPE_REASSOC_RESP = 0x30
IEEE80211_STYPE_PROBE_REQ = 0x40
IEEE80211_STYPE_PROBE_RESP = 0x50
IEEE80211_STYPE_BEACON = 0x80
IEEE80211_STYPE_ATIM = 0x90
IEEE80211_STYPE_DISASSOC = 0xA0
IEEE80211_STYPE_AUTH = 0xB0
IEEE80211_STYPE_DEAUTH = 0xC0
IEEE80211_STYPE_ACTION = 0xD0

WLAN_STATUS_SUCCESS = 0
WLAN_STATUS_UNSPECIFIED_FAILURE = 1
WLAN_STATUS_AP_UNABLE_TO_HANDLE_NEW_STA = 17

WLAN_REASON_UNSPECIFIED = 1

def SUITE(oui, id):
	return (oui << 8) | id

WLAN_CIPHER_SUITE_CCMP = 0x000FAC04

WLAN_AKM_SUITE_PSK = SUITE(0x000FAC, 2)

WLAN_AUTH_OPEN = 0

WLAN_EID_SSID = 0
WLAN_EID_SUPP_RATES = 1
WLAN_EID_DS_PARAMS = 3
WLAN_EID_SUPPORTED_CHANNELS = 36
WLAN_EID_HT_CAPABILITY = 45
WLAN_EID_RSN = 48
WLAN_EID_EXT_CAPABILITY = 127


Channels = {
	1: 2412,
	6: 2437,
	11: 2462,
	36: 5180,
	40: 5200,
	44: 5220,
	48: 5240
}

Frequencies = {v: k for k, v in Channels.items()}

def map_frequency(freq):
	return Frequencies[freq]

def is_valid_channel(channel):
	return channel in Channels


def encode_elements(elements):
	stream = streams.StreamOut("<")
	for id in sorted(elements):
		stream.u8(id)
		stream.u8(len(elements[id]))
		stream.write(elements[id])
	return stream.get()

def decode_elements(data):
	elements = {}
	stream = streams.StreamIn(data, "<")
	while not stream.eof():
		id = stream.u8()
		length = stream.u8()
		elements[id] = stream.read(length)
	return elements


class MACAddress:
	def __init__(self, address=None):
		if address is None:
			self.fields = [0] * 6
		elif isinstance(address, str):
			self.parse(address)
		elif isinstance(address, bytes):
			if len(address) != 6:
				raise ValueError("Invalid MAC address: %s" %address.hex())
			self.fields = list(address)
		elif isinstance(address, int):
			if address < 0 or address > 0xFFFFFFFFFFFF:
				raise ValueError("Invalid MAC address: %#x" %address)
			self.fields = [(address >> (40 - i * 8)) & 0xFF for i in range(6)]
		else:
			raise ValueError("Invalid MAC address: %s" %address)
	
	def __eq__(self, other):
		return self.fields == other.fields
	def __hash__(self):
		return hash(str(self))
	
	def __bytes__(self):
		return bytes(self.fields)

	def __str__(self):
		return ":".join("%02x" %value for value in self.fields)
	def __repr__(self):
		return "MACAddress('%s')" %self
	
	def encode(self): return bytes(self.fields)
	def decode(self, data): self.fields = list(data)
	
	def parse(self, text):
		fields = text.split(":")
		if len(fields) != 6:
			raise ValueError("Invalid MAC address: %s" %text)
		
		for field in fields:
			if len(field) != 2 or field[0] not in string.hexdigits or field[1] not in string.hexdigits:
				raise ValueError("Invalid MAC address: %s" %text)
		
		self.fields = [int(field, 16) for field in fields]


class SSIDElement:
	def __init__(self):
		self.ssid = None
	
	def encode(self):
		return self.ssid.encode()
	
	def decode(self, data):
		self.ssid = data.decode()


class SuppRatesElement:
	def __init__(self):
		self.supported_rates = None
	
	def encode(self):
		return bytes(self.supported_rates)
	
	def decode(self, data):
		self.supported_rates = list(data)


class DSParamsElement:
	def __init__(self):
		self.current_channel = None
	
	def encode(self):
		return bytes([self.current_channel])


class RSNElement:
	def __init__(self):
		self.group_cipher_suite = None
		self.pairwise_cipher_suites = None
		self.akm_suites = None
		self.capabilities = None
	
	def encode(self):
		stream = streams.StreamOut("<")
		stream.u16(1) # Version
		stream.u32_be(self.group_cipher_suite)
		stream.u16(len(self.pairwise_cipher_suites))
		stream.repeat(self.pairwise_cipher_suites, stream.u32_be)
		stream.u16(len(self.akm_suites))
		stream.repeat(self.akm_suites, stream.u32_be)
		stream.u16(self.capabilities)
		return stream.get()


class RadiotapFrame:
	def __init__(self):
		self.data = None
		
		self.mactime = None # Bit 0
		self.flags = None # Bit 1
		self.rate = None # Bit 2
		
		# Bit 3
		self.frequency = None
		self.channel_flags = None
	
	def encode(self):
		present = 0
		if self.mactime is not None: present |= 1
		if self.flags is not None: present |= 2
		if self.rate is not None: present |= 4
		if self.frequency is not None: present |= 8
		
		stream = streams.StreamOut("<")
		stream.u8(0) # Version
		stream.pad(1)
		stream.skip(2) # Length
		
		if present & 1:
			stream.align(8)
			stream.u64(self.mactime)
		if present & 2: stream.u8(self.flags)
		if present & 4: stream.u8(self.rate)
		if present & 8:
			stream.align(2)
			stream.u16(self.frequency)
			stream.u16(self.channel_flags)
		
		stream.align(8)
		
		length = stream.tell()
		stream.write(self.data)
		
		stream.seek(2)
		stream.u16(length)
		
		return stream.get()
	
	def decode(self, data):
		stream = streams.StreamIn(data, "<")
		
		version = stream.u8()
		if version != 0:
			raise ValueError("Radiotap header has invalid version number: %i" %version)
		
		stream.pad(1)
		
		length = stream.u16()
		
		shift = 0
		present = 0
		while True:
			value = stream.u32()
			present |= value << shift
			if not value & 0x80000000:
				break
			shift += 32
		
		if present & 1:
			stream.align(8)
			self.mactime = stream.u64()
		if present & 2: self.flags = stream.u8()
		if present & 4: self.rate = stream.u8()
		if present & 8:
			stream.align(2)
			self.frequency = stream.u16()
			self.channel_flags = stream.u16()
		
		if stream.tell() > length:
			raise ValueError("Radiotap header has wrong length field")
		
		stream.seek(length)
		self.data = stream.readall()


class MACHeader:
	def __init__(self):
		self.frame_control = 0
		self.duration = 0
		self.address1 = MACAddress()
		self.address2 = MACAddress()
		self.address3 = MACAddress()
		self.sequence_control = 0
	
	def encode(self):
		stream = streams.StreamOut("<")
		stream.u16(self.frame_control)
		stream.u16(self.duration)
		stream.write(self.address1.encode())
		stream.write(self.address2.encode())
		stream.write(self.address3.encode())
		stream.u16(self.sequence_control)
		return stream.get()
	
	def decode(self, data):
		stream = streams.StreamIn(data, "<")
		self.frame_control = stream.u16()
		self.duration = stream.u16()
		self.address1 = MACAddress(stream.read(6))
		self.address2 = MACAddress(stream.read(6))
		self.address3 = MACAddress(stream.read(6))
		self.sequence_control = stream.u16()


class AssociationRequest:
	def __init__(self):
		self.source = None
		self.target = None
		
		self.capability_information = None
		self.listen_interval = None
		self.elements = {}
	
	def decode(self, data):
		stream = streams.StreamIn(data, "<")
		
		header = MACHeader()
		header.decode(stream.read(24))
		if header.frame_control != IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ASSOC_REQ:
			raise ValueError("Frame is not an association request")
		
		self.target = header.address1
		self.source = header.address2
		
		self.capability_information = stream.u16()
		self.listen_interval = stream.u16()
		self.elements = decode_elements(stream.readall())


class AssociationResponse:
	def __init__(self):
		self.source = None
		self.target = None
		
		self.capability_information = None
		self.status_code = None
		self.aid = None
		
		self.elements = {}
		
	def encode(self):
		stream = streams.StreamOut("<")
		
		header = MACHeader()
		header.frame_control = IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ASSOC_RESP
		header.address1 = self.target
		header.address2 = self.source
		header.address3 = self.source
		stream.write(header.encode())
		
		stream.u16(self.capability_information)
		stream.u16(self.status_code)
		stream.u16(self.aid)
		
		stream.write(encode_elements(self.elements))
		return stream.get()


class ProbeRequest:
	def __init__(self):
		self.source = None
		self.elements = {}
	
	def decode(self, data):
		stream = streams.StreamIn(data, "<")
		
		header = MACHeader()
		header.decode(stream.read(24))
		if header.frame_control != IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_PROBE_REQ:
			raise ValueError("Frame is not a probe request")
		
		self.source = header.address2
		self.elements = decode_elements(stream.readall())


class ProbeResponse:
	def __init__(self):
		self.source = None
		self.target = None
		
		self.timestamp = 0
		self.beacon_interval = 0
		self.capability_information = 0
		
		self.elements = {}
	
	def encode(self):
		stream = streams.StreamOut("<")
		
		header = MACHeader()
		header.frame_control = IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_PROBE_RESP
		header.address1 = self.target
		header.address2 = self.source
		header.address3 = self.source
		stream.write(header.encode())
		
		stream.u64(self.timestamp)
		stream.u16(self.beacon_interval)
		stream.u16(self.capability_information)
		stream.write(encode_elements(self.elements))
		return stream.get()


class BeaconFrame:
	def __init__(self):
		self.source = None
		
		self.timestamp = 0
		self.beacon_interval = 0
		self.capability_information = 0
		
		self.elements = {}
		
	def encode(self):
		stream = streams.StreamOut("<")
		
		header = MACHeader()
		header.frame_control = IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_BEACON
		header.address1 = MACAddress("ff:ff:ff:ff:ff:ff")
		header.address2 = self.source
		header.address3 = self.source
		stream.write(header.encode())
		
		stream.u64(self.timestamp)
		stream.u16(self.beacon_interval)
		stream.u16(self.capability_information)
		stream.write(encode_elements(self.elements))
		return stream.get()


class DisassociationFrame:
	def __init__(self):
		self.source = None
		self.target = None
		self.reason = None
		self.elements = {}
	
	def decode(self, data):
		stream = streams.StreamIn(data, "<")
		
		header = MACHeader()
		header.decode(stream.read(24))
		if header.frame_control != IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_DISASSOC:
			raise ValueError("Frame is not a disassociation frame")
		
		self.target = header.address1
		self.source = header.address2
		self.reason = stream.u16()
		self.elements = decode_elements(stream.readall())


class AuthenticationFrame:
	def __init__(self):
		self.source = None
		self.target = None
		self.bssid = None
		
		self.algorithm = None
		self.sequence = None
		self.status_code = WLAN_STATUS_SUCCESS
		
		self.elements = {}
	
	def decode(self, data):
		stream = streams.StreamIn(data, "<")
		
		header = MACHeader()
		header.decode(stream.read(24))
		if header.frame_control != IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_AUTH:
			raise ValueError("Frame is not an authentication frame")
		
		self.target = header.address1
		self.source = header.address2
		self.bssid = header.address3
		
		self.algorithm = stream.u16()
		self.sequence = stream.u16()
		self.status_code = stream.u16()
		
		self.elements = decode_elements(stream.readall())
	
	def encode(self):
		stream = streams.StreamOut("<")
		
		header = MACHeader()
		header.frame_control = IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_AUTH
		header.address1 = self.target
		header.address2 = self.source
		header.address3 = self.bssid
		stream.write(header.encode())
		
		stream.u16(self.algorithm)
		stream.u16(self.sequence)
		stream.u16(self.status_code)
		stream.write(encode_elements(self.elements))
		return stream.get()


class DeauthenticationFrame:
	def __init__(self):
		self.source = None
		self.target = None
		self.bssid = None
		self.reason = None
		self.elements = {}
	
	def encode(self):
		stream = streams.StreamOut("<")
		
		header = MACHeader()
		header.frame_control = IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_DEAUTH
		header.address1 = self.target
		header.address2 = self.source
		header.address3 = self.bssid
		stream.write(header.encode())
		
		stream.u16(self.reason)
		stream.write(encode_elements(self.elements))
		return stream.get()


class ActionFrame:
	def __init__(self):
		self.source = None
		self.action = None
	
	def decode(self, data):
		stream = streams.StreamIn(data, "<")
		
		header = MACHeader()
		header.decode(stream.read(24))
		if header.frame_control != IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ACTION:
			raise ValueError("Frame is not an action frame")
		
		self.source = header.address2
		
		self.action = stream.readall()
	
	def encode(self):
		header = MACHeader()
		header.frame_control = IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ACTION
		header.address1 = MACAddress("ff:ff:ff:ff:ff:ff")
		header.address2 = self.source
		header.address3 = MACAddress("ff:ff:ff:ff:ff:ff")
		
		stream = streams.StreamOut("<")
		stream.write(header.encode())
		stream.write(self.action)
		return stream.get()


class AssociationEvent:
	def __init__(self, address):
		self.address = address
		

class DisassociationEvent:
	def __init__(self, address):
		self.address = address


class FrameEvent:
	def __init__(self, frame, frequency):
		self.frame = frame
		self.frequency = frequency


class DataFrameEvent:
	def __init__(self, address, data):
		self.address = address
		self.data = data


class Interface:
	def __init__(self, wlan, attributes):
		self.wlan = wlan
		
		self.phy = attributes[nl80211.NL80211_ATTR_WIPHY]
		self.index = attributes[nl80211.NL80211_ATTR_IFINDEX]
		self.name = attributes[nl80211.NL80211_ATTR_IFNAME]
		self.type = attributes[nl80211.NL80211_ATTR_IFTYPE]
		self.address = MACAddress(attributes[nl80211.NL80211_ATTR_MAC])
		
		self.socket = socket.socket()
	
	def getflags(self):
		req = struct.pack("16sH", self.name.encode(), 0)
		res = fcntl.ioctl(self.socket.fileno(), SIOCGIFFLAGS, req)
		return struct.unpack_from("H", res, 16)[0]
	
	def setflags(self, flags):
		# This is really slow, but there doesn't seem to be
		# an async implementation of fcntl.ioctl :(
		req = struct.pack("16sH", self.name.encode(), flags)
		fcntl.ioctl(self.socket.fileno(), SIOCSIFFLAGS, req)
	
	def disable_ipv6(self):
		with open("/proc/sys/net/ipv6/conf/%s/disable_ipv6" %self.name, "w") as f:
			f.write("1")
	
	def up(self):
		self.setflags(self.getflags() | IFF_UP)
		
	async def set_channel(self, channel):
		if channel not in Channels:
			raise ValueError("Invalid channel: %i" %channel)
		
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.index,
			nl80211.NL80211_ATTR_WIPHY_FREQ: Channels[channel]
		}
		await self.wlan.request(nl80211.NL80211_CMD_SET_CHANNEL, attrs)


class Monitor:
	# This class represents an interface in monitor mode
	
	def __init__(self, wlan, interface):
		self.wlan = wlan
		self.interface = interface
		
		self.socket = trio.socket.socket(socket.AF_PACKET, socket.SOCK_RAW, socket.htons(ETH_P_ALL))
	
	async def bind(self):
		await self.socket.bind((self.interface.name, 0))
	
	async def recv(self):
		while True:
			data = await self.socket.recv(4096)
			radiotap = RadiotapFrame()
			try:
				radiotap.decode(data)
				return radiotap
			except Exception as e:
				pass # Ignore invalid frames
	
	async def send(self, frame):
		await self.socket.send(frame.encode())
	
	async def set_channel(self, channel):
		await self.interface.set_channel(channel)


class STAInterface:
	# This class represents an interface in station mode
	
	def __init__(self, wlan, interface, ssid, channel, key):
		self.wlan = wlan
		self.interface = interface
		self.ssid = ssid
		self.channel = channel
		self.key = key
		
		self.address = self.interface.address
		self.index = self.interface.index
		
		self.host_address = None
		
		self.events = queue.create()
	
	async def next_event(self):
		return await self.events.get()
	
	async def send_data_frame(self, addr, frame):
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_FRAME: frame,
			nl80211.NL80211_ATTR_MAC: addr.encode(),
			nl80211.NL80211_ATTR_CONTROL_PORT_ETHERTYPE: struct.pack("H", ETH_P_OUI)
		}
		await self.wlan.request(nl80211.NL80211_CMD_CONTROL_PORT_FRAME, attrs)
	
	async def register_keys(self):
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_MAC: self.host_address,
			nl80211.NL80211_ATTR_KEY: {
				nl80211.NL80211_KEY_IDX: 0,
				nl80211.NL80211_KEY_DATA: self.key,
				nl80211.NL80211_KEY_CIPHER: WLAN_CIPHER_SUITE_CCMP
			}
		}
		await self.wlan.request(nl80211.NL80211_CMD_NEW_KEY, attrs)
		
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_KEY: {
				nl80211.NL80211_KEY_IDX: 1,
				nl80211.NL80211_KEY_DATA: self.key,
				nl80211.NL80211_KEY_CIPHER: WLAN_CIPHER_SUITE_CCMP
			}
		}
		await self.wlan.request(nl80211.NL80211_CMD_NEW_KEY, attrs)

	@contextlib.asynccontextmanager
	async def connect_network(self):
		rsn = RSNElement()
		rsn.group_cipher_suite = WLAN_CIPHER_SUITE_CCMP
		rsn.pairwise_cipher_suites = [WLAN_CIPHER_SUITE_CCMP]
		rsn.akm_suites = [WLAN_AKM_SUITE_PSK]
		rsn.capabilities = 0xC
		
		elements = {
			WLAN_EID_RSN: rsn.encode()
		}
		
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_SSID: self.ssid.encode(),
			nl80211.NL80211_ATTR_WIPHY_FREQ: Channels[self.channel],
			nl80211.NL80211_ATTR_AUTH_TYPE: nl80211.NL80211_AUTHTYPE_OPEN_SYSTEM,
			nl80211.NL80211_ATTR_CIPHER_SUITES_PAIRWISE: struct.pack("I", WLAN_CIPHER_SUITE_CCMP),
			nl80211.NL80211_ATTR_CIPHER_SUITE_GROUP: WLAN_CIPHER_SUITE_CCMP,
			nl80211.NL80211_ATTR_AKM_SUITES: struct.pack("I", WLAN_AKM_SUITE_PSK),
			nl80211.NL80211_ATTR_IE: encode_elements(elements),
			nl80211.NL80211_ATTR_PRIVACY: True,
			
			nl80211.NL80211_ATTR_CONTROL_PORT: True,
			nl80211.NL80211_ATTR_CONTROL_PORT_ETHERTYPE: struct.pack("H", ETH_P_OUI),
			nl80211.NL80211_ATTR_CONTROL_PORT_OVER_NL80211: True,
			nl80211.NL80211_ATTR_SOCKET_OWNER: True
		}
		await self.wlan.request(nl80211.NL80211_CMD_CONNECT, attrs)
		
		while True:
			message = await self.wlan.receive()
			if message.type == nl80211.NL80211_CMD_CONNECT:
				status = message.attributes[nl80211.NL80211_ATTR_STATUS_CODE]
				if status != WLAN_STATUS_SUCCESS:
					raise ConnectionError("Connect failed with status code %i" %status)
				break
		
		try:
			self.host_address = message.attributes[nl80211.NL80211_ATTR_MAC]
			if self.key is not None:
				await self.register_keys()
			yield
		finally:
			attrs = {nl80211.NL80211_ATTR_IFINDEX: self.interface.index}
			await self.wlan.request(nl80211.NL80211_CMD_DISCONNECT, attrs)
	
	@contextlib.asynccontextmanager
	async def connect(self):
		self.interface.disable_ipv6()
		async with self.connect_network() as addr:
			async with util.background_task(self.process_messages):
				await self.register_frame(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ACTION)
				yield
	
	async def register_frame(self, type, match=b""):
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_FRAME_TYPE: type,
			nl80211.NL80211_ATTR_FRAME_MATCH: match
		}
		await self.wlan.request(nl80211.NL80211_CMD_REGISTER_FRAME, attrs)
	
	async def process_messages(self):
		while True:
			message = await self.wlan.receive()
			if message.type == nl80211.NL80211_CMD_FRAME:
				frame = ActionFrame()
				frame.decode(message.attributes[nl80211.NL80211_ATTR_FRAME])
				freq = message.attributes[nl80211.NL80211_ATTR_WIPHY_FREQ]
				await self.events.put(FrameEvent(frame, freq))
			elif message.type == nl80211.NL80211_CMD_CONTROL_PORT_FRAME:
				address = MACAddress(message.attributes[nl80211.NL80211_ATTR_MAC])
				data = message.attributes[nl80211.NL80211_ATTR_FRAME]
				await self.events.put(DataFrameEvent(address, data))
			elif message.type == nl80211.NL80211_CMD_DEL_STATION:
				address = MACAddress(message.attributes[nl80211.NL80211_ATTR_MAC])
				await self.events.put(DisassociationEvent(None))
	
	async def set_authorized(self):
		flag = 1 << nl80211.NL80211_STA_FLAG_AUTHORIZED
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_MAC: self.host_address,
			nl80211.NL80211_ATTR_STA_FLAGS2: struct.pack("II", flag, flag)
		}
		await self.wlan.request(nl80211.NL80211_CMD_SET_STATION, attrs)


class APInterface:
	# This class represents an interface in access point mode
	
	def __init__(self, wlan, interface, ssid, channel, key, max_stations):
		self.wlan = wlan
		self.interface = interface
		self.ssid = ssid
		self.channel = channel
		self.key = key
		self.max_stations = max_stations
		
		self.address = self.interface.address
		self.index = self.interface.index
		
		self.stations_by_id = {}
		self.stations_by_address = {}
		
		self.events = queue.create()
	
	def create_beacon_head(self):
		frame = BeaconFrame()
		frame.source = self.interface.address
		frame.beacon_interval = 100
		frame.capability_information = 0x511
		return frame.encode()
	
	def create_beacon_tail(self):
		return b"" # No beacon tail for now
	
	def create_probe_response(self, address):
		ssid = SSIDElement()
		ssid.ssid = self.ssid
		
		rates = SuppRatesElement()
		rates.supported_rates = [0x82, 0x84, 0x8B, 0x96, 0x24, 0x30, 0x48, 0x6C]
		
		dsparams = DSParamsElement()
		dsparams.current_channel = self.channel
		
		rsn = RSNElement()
		rsn.group_cipher_suite = WLAN_CIPHER_SUITE_CCMP
		rsn.pairwise_cipher_suites = [WLAN_CIPHER_SUITE_CCMP]
		rsn.akm_suites = [WLAN_AKM_SUITE_PSK]
		rsn.capabilities = 0xC
		
		response = ProbeResponse()
		response.source = self.interface.address
		response.target = address
		response.beacon_interval = 100
		response.capability_information = 0x511
		response.elements = {
			WLAN_EID_SSID: ssid.encode(),
			WLAN_EID_SUPP_RATES: rates.encode(),
			WLAN_EID_DS_PARAMS: dsparams.encode(),
			WLAN_EID_RSN: rsn.encode()
		}
		return response.encode()
	
	def create_association_response(self, address, aid):
		rates = SuppRatesElement()
		rates.supported_rates = [0x82, 0x84, 0x8B, 0x96, 0x24, 0x30, 0x48, 0x6C]
		
		response = AssociationResponse()
		response.source = self.interface.address
		response.target = address
		response.capability_information = 0x411
		response.status_code = WLAN_STATUS_SUCCESS
		response.aid = aid | 0xC000
		response.elements = {
			WLAN_EID_SUPP_RATES: rates.encode()
		}
		return response.encode()
	
	def create_association_error(self, address, error):
		response = AssociationResponse()
		response.source = self.interface.address
		response.target = address
		response.capability_information = 0x411
		response.status_code = error
		response.aid = 0
		return response.encode()
	
	def parse_management_frame(self, data):
		header = MACHeader()
		header.decode(data)
		
		frame = {
			IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ASSOC_REQ: AssociationRequest,
			IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_PROBE_REQ: ProbeRequest,
			IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_DISASSOC: DisassociationFrame,
			IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_AUTH: AuthenticationFrame
		}[header.frame_control]()
		frame.decode(data)
		return frame
	
	async def next_event(self):
		return await self.events.get()
	
	@contextlib.asynccontextmanager
	async def start_ap(self):
		beacon_head = self.create_beacon_head()
		beacon_tail = self.create_beacon_tail()
		
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_SSID: self.ssid.encode(),
			nl80211.NL80211_ATTR_BEACON_HEAD: beacon_head,
			nl80211.NL80211_ATTR_BEACON_TAIL: beacon_tail,
			nl80211.NL80211_ATTR_BEACON_INTERVAL: 100,
			nl80211.NL80211_ATTR_DTIM_PERIOD: 3,
			nl80211.NL80211_ATTR_HIDDEN_SSID: nl80211.NL80211_HIDDEN_SSID_ZERO_CONTENTS,
			nl80211.NL80211_ATTR_WIPHY_FREQ: Channels[self.channel],
			nl80211.NL80211_ATTR_CONTROL_PORT: True,
			nl80211.NL80211_ATTR_CONTROL_PORT_ETHERTYPE: struct.pack("H", ETH_P_OUI),
			nl80211.NL80211_ATTR_CONTROL_PORT_OVER_NL80211: True,
			nl80211.NL80211_ATTR_SOCKET_OWNER: True
			
		}
		await self.wlan.request(nl80211.NL80211_CMD_START_AP, attrs)
		
		if self.key is not None:
			attrs = {
				nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
				nl80211.NL80211_ATTR_KEY: {
					nl80211.NL80211_KEY_IDX: 1,
					nl80211.NL80211_KEY_DATA: self.key,
					nl80211.NL80211_KEY_CIPHER: WLAN_CIPHER_SUITE_CCMP
				}
			}
			await self.wlan.request(nl80211.NL80211_CMD_NEW_KEY, attrs)

			attrs = {
				nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
				nl80211.NL80211_ATTR_KEY: {
					nl80211.NL80211_KEY_IDX: 1,
					nl80211.NL80211_KEY_DEFAULT: True,
					nl80211.NL80211_KEY_DEFAULT_TYPES: {
						nl80211.NL80211_KEY_DEFAULT_TYPE_MULTICAST: True
					}
				}
			}
			await self.wlan.request(nl80211.NL80211_CMD_SET_KEY, attrs)
		
		try:
			yield
		finally:
			attrs = {nl80211.NL80211_ATTR_IFINDEX: self.interface.index}
			await self.wlan.request(nl80211.NL80211_CMD_STOP_AP, attrs)
	
	@contextlib.asynccontextmanager
	async def start(self):
		self.interface.disable_ipv6()
		async with self.start_ap():
			await self.register_frame(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ASSOC_REQ)
			await self.register_frame(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_PROBE_REQ)
			await self.register_frame(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_DISASSOC)
			await self.register_frame(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_AUTH)
			
			async with util.background_task(self.process_messages):
				yield
	
	async def register_frame(self, type, match=b""):
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_FRAME_TYPE: type,
			nl80211.NL80211_ATTR_FRAME_MATCH: match
		}
		await self.wlan.request(nl80211.NL80211_CMD_REGISTER_FRAME, attrs)
	
	async def process_messages(self):
		while True:
			message = await self.wlan.receive()
			if message.type == nl80211.NL80211_CMD_FRAME:
				data = message.attributes[nl80211.NL80211_ATTR_FRAME]
				try:
					frame = self.parse_management_frame(data)
				except Exception as e:
					continue # Ignore invalid frames
				await self.process_frame(frame)
			elif message.type == nl80211.NL80211_CMD_CONTROL_PORT_FRAME:
				address = MACAddress(message.attributes[nl80211.NL80211_ATTR_MAC])
				data = message.attributes[nl80211.NL80211_ATTR_FRAME]
				await self.events.put(DataFrameEvent(address, data))
	
	async def process_frame(self, frame):
		if isinstance(frame, ProbeRequest):
			ssid = frame.elements.get(WLAN_EID_SSID)
			if ssid == self.ssid.encode():
				response = self.create_probe_response(frame.source)
				await self.send_frame(response)
		elif isinstance(frame, AuthenticationFrame):
			if frame.bssid == self.interface.address:
				if frame.algorithm == WLAN_AUTH_OPEN and frame.sequence == 1:
					response = AuthenticationFrame()
					response.source = self.interface.address
					response.target = frame.source
					response.bssid = self.interface.address
					response.algorithm = WLAN_AUTH_OPEN
					response.sequence = 2
					response.status_code = WLAN_STATUS_SUCCESS
					await self.send_frame(response.encode())
		elif isinstance(frame, AssociationRequest):
			ssid = frame.elements.get(WLAN_EID_SSID)
			if ssid == self.ssid.encode():
				response = await self.process_association_request(frame)
				await self.send_frame(response)
		elif isinstance(frame, DisassociationFrame):
			await self.process_disassociation(frame)
	
	async def process_association_request(self, frame):
		if frame.source in self.stations_by_address:
			aid = self.stations_by_address[frame.source]
			return self.create_association_response(frame.source, aid)
		
		if len(self.stations_by_id) >= self.max_stations:
			return self.create_association_error(frame.source, WLAN_STATUS_AP_UNABLE_TO_HANDLE_NEW_STA)
		
		if WLAN_EID_SUPP_RATES not in frame.elements:
			return self.create_association_error(frame.source, WLAN_STATUS_ASSOC_DENIED_UNSPEC)
		
		aid = 1
		while aid in self.stations_by_id:
			aid += 1
		
		self.stations_by_id[aid] = frame.source
		self.stations_by_address[frame.source] = aid
		
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_MAC: frame.source.encode(),
			nl80211.NL80211_ATTR_STA_LISTEN_INTERVAL: frame.listen_interval,
			nl80211.NL80211_ATTR_STA_SUPPORTED_RATES: frame.elements[WLAN_EID_SUPP_RATES],
			nl80211.NL80211_ATTR_STA_CAPABILITY: frame.capability_information,
			nl80211.NL80211_ATTR_STA_AID: aid
		}
		if WLAN_EID_EXT_CAPABILITY in frame.elements:
			attrs[nl80211.NL80211_ATTR_STA_EXT_CAPABILITY] = frame.elements[WLAN_EID_EXT_CAPABILITY]
		if WLAN_EID_HT_CAPABILITY in frame.elements:
			attrs[nl80211.NL80211_ATTR_HT_CAPABILITY] = frame.elements[WLAN_EID_HT_CAPABILITY]
		if WLAN_EID_SUPPORTED_CHANNELS in frame.elements:
			attrs[nl80211.NL80211_ATTR_STA_SUPPORTED_CHANNELS] = frame.elements[WLAN_EID_SUPPORTED_CHANNELS]
		await self.wlan.request(nl80211.NL80211_CMD_NEW_STATION, attrs)
		
		if self.key is not None:
			attrs = {
				nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
				nl80211.NL80211_ATTR_MAC: frame.source.encode(),
				nl80211.NL80211_ATTR_KEY: {
					nl80211.NL80211_KEY_IDX: 0,
					nl80211.NL80211_KEY_DATA: self.key,
					nl80211.NL80211_KEY_CIPHER: WLAN_CIPHER_SUITE_CCMP
				}
			}
			await self.wlan.request(nl80211.NL80211_CMD_NEW_KEY, attrs)
		
		await self.events.put(AssociationEvent(frame.source))
		return self.create_association_response(frame.source, aid)
	
	async def process_disassociation(self, frame):
		if frame.source not in self.stations_by_address: return
		
		aid = self.stations_by_address.pop(frame.source)
		del self.stations_by_id[aid]
		
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_MAC: frame.source.encode(),
			nl80211.NL80211_ATTR_MGMT_SUBTYPE: IEEE80211_STYPE_DISASSOC >> 4,
			nl80211.NL80211_ATTR_REASON_CODE: frame.reason
		}
		await self.wlan.request(nl80211.NL80211_CMD_DEL_STATION, attrs)
		
		await self.events.put(DisassociationEvent(frame.source))
	
	async def send_frame(self, data):
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_FRAME: data
		}
		await self.wlan.request(nl80211.NL80211_CMD_FRAME, attrs)
	
	async def send_data_frame(self, addr, frame):
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_FRAME: frame,
			nl80211.NL80211_ATTR_MAC: addr.encode(),
			nl80211.NL80211_ATTR_CONTROL_PORT_ETHERTYPE: struct.pack("H", ETH_P_OUI)
		}
		await self.wlan.request(nl80211.NL80211_CMD_CONTROL_PORT_FRAME, attrs)
	
	async def set_authorized(self, addr):
		flag = 1 << nl80211.NL80211_STA_FLAG_AUTHORIZED
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_MAC: addr.encode(),
			nl80211.NL80211_ATTR_STA_FLAGS2: struct.pack("II", flag, flag)
		}
		await self.wlan.request(nl80211.NL80211_CMD_SET_STATION, attrs)
	
	async def remove_station(self, addr):
		if addr not in self.stations_by_address: return
		
		aid = self.stations_by_address.pop(addr)
		del self.stations_by_id[aid]
		
		frame = DeauthenticationFrame()
		frame.source = self.interface.address
		frame.target = addr
		frame.bssid = self.interface.address
		frame.reason = WLAN_REASON_UNSPECIFIED
		await self.send_management_frame(frame.encode())
		
		attrs = {
			nl80211.NL80211_ATTR_IFINDEX: self.interface.index,
			nl80211.NL80211_ATTR_MAC: addr.encode(),
			nl80211.NL80211_ATTR_REASON_CODE: WLAN_REASON_UNSPECIFIED
		}
		await self.wlan.request(nl80211.NL80211_CMD_DEL_STATION, attrs)


class WLAN:
	# This class acts as a factory for WLAN interfaces
	
	def __init__(self, wlan):
		self.wlan = wlan
		self.wlan.add_membership("mlme")
		
	async def get_wiphy_index(self, name):
		messages = await self.wlan.request(nl80211.NL80211_CMD_GET_WIPHY, flags=netlink.NLM_F_DUMP)
		for message in messages:
			if message.attributes[nl80211.NL80211_ATTR_WIPHY_NAME] == name:
				return message.attributes[nl80211.NL80211_ATTR_WIPHY]
		raise ValueError("No wiphy found with name '%s'" %name)

	@contextlib.asynccontextmanager
	async def create_interface(self, phy, name, type, extra={}):
		index = await self.get_wiphy_index(phy)
		attrs = {
			nl80211.NL80211_ATTR_WIPHY: index,
			nl80211.NL80211_ATTR_IFNAME: name,
			nl80211.NL80211_ATTR_IFTYPE: type,
		}
		attrs.update(extra)
		
		messages = await self.wlan.request(nl80211.NL80211_CMD_NEW_INTERFACE, attrs)
		interface = Interface(self.wlan, messages[0].attributes)
		try:
			interface.up()
			yield interface
		finally:
			attrs = {nl80211.NL80211_ATTR_IFINDEX: interface.index}
			await self.wlan.request(nl80211.NL80211_CMD_DEL_INTERFACE, attrs)
	
	@contextlib.asynccontextmanager
	async def create_monitor(self, phy, name):
		attrs = {
			nl80211.NL80211_ATTR_MNTR_FLAGS: {
				nl80211.NL80211_MNTR_FLAG_OTHER_BSS: True
			}
		}
		async with self.create_interface(phy, name, nl80211.NL80211_IFTYPE_MONITOR, attrs) as interface:
			monitor = Monitor(self.wlan, interface)
			await monitor.bind()
			yield monitor
	
	@contextlib.asynccontextmanager
	async def connect_network(self, phy, name, ssid, channel, key):
		async with self.create_interface(phy, name, nl80211.NL80211_IFTYPE_STATION) as interface:
			interface = STAInterface(self.wlan, interface, ssid, channel, key)
			async with interface.connect():
				yield interface
	
	@contextlib.asynccontextmanager
	async def create_network(self, phy, name, ssid, channel, key, max_stations):
		async with self.create_interface(phy, name, nl80211.NL80211_IFTYPE_AP) as interface:
			interface = APInterface(self.wlan, interface, ssid, channel, key, max_stations)
			async with interface.start():
				yield interface


@contextlib.asynccontextmanager
async def create():
	async with nl80211.connect() as wlan:
		yield WLAN(wlan)




import setuptools

description = \
	"Python package for local wireless communication with a Nintendo Switch."

long_description = \
	"This package implements LDN, or local wireless communication " \
	"with a Nintendo Switch. Because LDN operates at the data link " \
	"layer it requires low-level access to your network hardware. " \
	"This package only supports Linux systems."

setuptools.setup(
	name = "ldn",
	version = "0.0.10",
	description = description,
	long_description = long_description,
	author = "Yannik Marchand",
	author_email = "ymarchand@me.com",
	url = "https://github.com/kinnay/LDN",
	license = "GPLv3",
	platforms = ["Linux"],
	packages = ["ldn"],
	install_requires = [
		"python-netlink == 0.0.9",
		"pycryptodome"
	]
)




import math
import trio


class Queue:
	def __init__(self, sender, receiver):
		self.sender = sender
		self.receiver = receiver
	
	async def __aenter__(self): return self
	async def __aexit__(self, typ, val, tb):
		await self.close()
	
	async def put(self, value):
		await self.sender.send(value)
	
	async def get(self):
		return await self.receiver.receive()
	
	async def close(self):
		await self.sender.aclose()
		await self.receiver.aclose()


def create(size=math.inf):
	send, recv = trio.open_memory_channel(size)
	return Queue(send, recv)
